/*
 * Copyright (C) 2001-2002 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *
 *    This product includes software developed by egnite Software GmbH
 *    and its contributors.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 *
 */

/*
 * $Log$
 */

#include <sys/ifdisplay.h>

#ifdef NUTDEBUG
#include <sys/print.h>
#endif

/*!
 * \addtogroup xgIfDisplay
 */
/*@{*/


/*!
 * \brief Write to a display device.
 *
 * The data isn't immediately transfered to the physical device, 
 * but buffered in a transmit buffer. Transmission starts either 
 * when the buffer is full or when this function is called with
 * a zero data pointer.
 *
 * \param dev  Identifies the device to write to. This pointer
 *             must have been retrieved by calling NutDeviceOpen().
 * \param data Pointer to data to be written.
 * \param len  Number of bytes to write.
 *
 * \return The number of bytes written or -1 in case of an error.
 */
int NutIfDisplayWrite(NUTDEVICE *dev, CONST char *data, int len)
{
    int rc;
    IFDISPLAY *ifs;
    u_char ix;

    /*
     * Check for valid device pointer.
     */
    if((void *)dev < RAMSTART)
        return -1;

    /*
     * Check if output is supported.
     */
    ifs = dev->dev_icb;
    if(ifs->if_output == 0)
        return -1;

    /*
     * Call without data pointer starts transmission.
     */
    if(data == 0) {
        if((*ifs->if_output)(dev) < 0)
            return -1;
        return 0;
    }

    /*
     * Check for valid buffer pointer.
     */
    if((void *)data < RAMSTART)
        return -1;

    /*
     * Start transmission if output buffer full.
     */
    if((ix = ifs->if_wr_idx + 1) == ifs->if_tx_idx)
        if((*ifs->if_output)(dev) < 0)
            return -1;

    /*
     * Put characters in transmit buffer.
     */
    for(rc = 0; rc < len; rc++) {
        if(ix == ifs->if_tx_idx)
            break;
        ifs->if_tx_buf[ifs->if_wr_idx] = *data++;
        ifs->if_wr_idx = ix++;
    }
    return rc;
}


/*!
 * \brief Write program space data to a display device.
 *
 * \param dev  Identifies the device to write to. This pointer
 *             must have been retrieved by calling NutDeviceOpen().
 * \param data Pointer to data in program space to be written.
 * \param len  Number of bytes to write.
 *
 * \return The number of bytes written or -1 in case of an error.
 */
int NutIfDisplayWrite_P(NUTDEVICE *dev, PGM_P data, int len)
{
    int rc;
    IFDISPLAY *ifs;
    u_char ix;

    /*
     * Check for valid device pointer.
     */
    if((void *)dev < RAMSTART)
        return -1;

    /*
     * Check if output is supported.
     */
    ifs = (IFDISPLAY *)dev->dev_icb;
    if(ifs->if_output == 0)
        return -1;

    /*
     * If buffer full, wait for output.
     */
    if((ix = ifs->if_wr_idx + 1) == ifs->if_tx_idx)
        if((*ifs->if_output)(dev) < 0)
            return -1;

    /*
     * Put characters in transmit buffer.
     */
    for(rc = 0; rc < len; rc++) {
        if(ix == ifs->if_tx_idx)
            break;
        ifs->if_tx_buf[ifs->if_wr_idx] = PRG_RDB(data);
        data++;
        ifs->if_wr_idx = ix++;
    }
    return rc;
}


/*@}*/
